# encoding: utf-8
"""
Function align is used by DataView, to expand and re-arrange data in a DataFrame
according to their available time, which is stored in another DataFrame.

"""
from __future__ import print_function
import numpy as np
import pandas as pd
from jaqs.util import is_numeric


def _get_neareast(df_ann, df_value, date):
    """
    Get the value whose ann_date is earlier and nearest to date.
    
    Parameters
    ----------
    df_ann : np.ndarray
        announcement dates. shape = (n_quarters, n_securities)
    df_value : np.ndarray
        announcement values. shape = (n_quarters, n_securities)
    date : np.ndarray
        shape = (1,)

    Returns
    -------
    res : np.array
        The value whose ann_date is earlier and nearest to date. shape (n_securities)

    """
    """
    df_ann.fillna(99999999, inplace=True)  # IMPORTANT: At cells where no quarterly data is available,
                                           # we know nothing, thus it will be filled nan in the next step
    """
    if is_numeric(df_value):
        df_value = df_value.astype(float)
    mask = date[0] >= df_ann
    # res = np.where(mask, df_value, np.nan)
    n = df_value.shape[1]
    res = np.empty(n, dtype=df_value.dtype)
    
    # for each column, get the last True value
    for i in range(n):
        v = df_value[:, i]
        m = mask[:, i]
        r = v[m]
        res[i] = r[-1] if len(r) else np.nan
    
    return res
    

def align(df_value, df_ann, date_arr):
    """
    Expand low frequency DataFrame df_value to frequency of data_arr using announcement date from df_ann.
    
    Parameters
    ----------
    df_ann : pd.DataFrame
        DataFrame of announcement dates. shape = (n_quarters, n_securities)
    df_value : pd.DataFrame
        DataFrame of announcement values. shape = (n_quarters, n_securities)
    date_arr : list or np.array
        Target date array. dtype = int

    Returns
    -------
    df_res : pd.DataFrame
        Expanded DataFrame. shape = (n_days, n_securities)

    """
    df_ann = df_ann.fillna(99999999).astype(int)
    
    date_arr = np.asarray(date_arr, dtype=int)
    
    res = np.apply_along_axis(lambda date: _get_neareast(df_ann.values, df_value.values, date), 1, date_arr.reshape(-1, 1))

    df_res = pd.DataFrame(index=date_arr, columns=df_value.columns, data=res)
    return df_res



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import json
import pandas as pd
from . import utils


class EntrustOrder(object):
    def __init__(self, security, action, price, size):
        self.security  = security
        self.action    = action
        self.price     = price
        self.size      = size


def set_log_dir(log_dir):
    try:
        import jrpc
        if log_dir:
            jrpc.set_log_dir(log_dir)
        else:
            jrpc.set_log_dir("")
    except Exception as e:
        print("Exception", e)


class TradeApi(object):
    def __init__(self, addr, use_jrpc=False, prod_type="jzts"):
        """
            use_jrpc:
                True     -- Use jrcp_client of C version, for jzts only
                False    -- Use pure python version
            prod_type:
                "jaqs"   -- jrpc_msgpack_wth_snappy
                "jzts"   -- jrpc_msgpack
        """

        self._remote = None
        if prod_type == "jzts":
            try:
                if use_jrpc:
                    import jrpc
                    self._remote = jrpc.JsonRpcClient()
                else:
                    from . import jrpc_py
                    self._remote = jrpc_py.JRpcClient(data_format="msgpack")
            except Exception as e:
                print("Exception", e)

            if not self._remote:
                from . import jrpc_py
                self._remote = jrpc_py.JRpcClient(data_format="msgpack")

        else:
            from . import jrpc_py
            self._remote = jrpc_py.JRpcClient(data_format="msgpack_snappy")

        self._remote.on_rpc_callback = self._on_rpc_callback
        self._remote.on_disconnected = self._on_disconnected
        self._remote.on_connected    = self._on_connected
        self._remote.connect(addr)

        self._ordstatus_callback = None
        self._taskstatus_callback = None
        self._internal_order_callback = None
        self._trade_callback = None
        self._on_connection_callback = None
        self._connected   = False
        self._username    = ""
        self._password    = ""
        self._strategy_id = 0
        self._strategy_selected = False
        self._data_format = "default"

    def __del__(self):
        self._remote.close()
    
    def _on_rpc_callback(self, method, data):
        #print "_on_rpc_callback:", method, data

        if method == "oms.orderstatus_ind":
            if self._data_format == "obj":
                data = utils.to_obj("Order", data)

            if self._ordstatus_callback:
                self._ordstatus_callback(data)

        elif method == "oms.taskstatus_ind":
            if self._data_format == "obj":
                data = utils.to_obj("TaskStatus", data)

            if self._taskstatus_callback:
                self._taskstatus_callback(data)

        elif method == "oms.trade_ind":
            if self._data_format == "obj":
                data = utils.to_obj("Trade", data)

            if self._trade_callback:
                self._trade_callback(data)

        elif method == "oms.internal_order_ind":
            if self._data_format == "obj":
                data = utils.to_obj("QuoteOrder", data)

            if self._internal_order_callback:
                self._internal_order_callback(data)

    def _on_disconnected(self):
        print("TradeApi: _on_disconnected")
        self._connected = False
        self._strategy_selected = False
        if self._on_connection_callback:
            self._on_connection_callback(False)

    def _on_connected(self):
        print("TradeApi: _on_connected")
        self._connected = True
        self._do_login()
        self._do_use_strategy()
        if self._on_connection_callback:
            self._on_connection_callback(True)

    def _check_session(self):
        if not self._connected:
            return (False, "no connection")

        if self._strategy_selected:
            return (True, "")

        r, msg = self._do_login()
        if not r: return (r, msg)
        if self._strategy_id:
            return self._do_use_strategy()
        else:
            return (r, msg)

    def set_data_format(self, format):
        self._data_format = format

    def set_connection_callback(self, callback):
        self._on_connection_callback = callback

    def set_ordstatus_callback(self, callback):
        self._ordstatus_callback = callback

    def set_trade_callback(self, callback):
        self._trade_callback = callback

    def set_task_callback(self, callback):
        self._taskstatus_callback = callback

    def set_quoteorder_callback(self, callback):
        self._internal_order_callback = callback

    def _get_format(self, format, default_format):
        if format:
            return format
        elif self._data_format != "default":
            return self._data_format
        else:
            return default_format

    def login(self, username, password, format=""):
        self._username = username
        self._password = password
        return self._do_login(format=format)

    def _do_login(self, format=""):
        # Shouldn't check connected flag here. ZMQ is a mesageq queue!
        # if !self._connected :
        #    return (False, "-1,no connection")

        if self._username and self._password:
            rpc_params = {"username": self._username,
                          "password": self._password}

            cr = self._remote.call("auth.login", rpc_params)
            f = self._get_format(format, "")
            if f != "obj" and f != "":
                f = ""
            return utils.extract_result(cr, data_format=f, class_name="UserInfo")
        else:
            return (False, "-1,empty username or password")

    def logout(self):
        rpc_params = {}
        
        cr = self._remote.call("auth.logout", rpc_params)
        return utils.extract_result(cr)

    def close(self):
        self._remote.close()

    def use_strategy(self, strategy_id):
        if strategy_id:
            self._strategy_id = strategy_id
            return self._do_use_strategy()
        else:
            # Query
            rpc_params = {"account_id": 0}

            cr = self._remote.call("auth.use_strategy", rpc_params)
            r, msg = utils.extract_result(cr)
            self._strategy_selected = r

            return (r, msg)

    def _do_use_strategy(self):
        if self._strategy_id:
            rpc_params = {"account_id": self._strategy_id}

            cr = self._remote.call("auth.use_strategy", rpc_params)
            r, msg = utils.extract_result(cr)
            self._strategy_selected = r

            return (r, msg)
        else:
            return (False, "-1,no strategy_id was specified")

    def confirm_internal_order(self, task_id, confirmed):
        """
        return (result, message)
        if result is None, message contains error information
        """

        r, msg = self._check_session()
        if not r: return (None, msg)

        rpc_params = {"task_id"  : task_id,
                      "confirmed": confirmed}

        cr = self._remote.call("oms.confirm_internal_order", rpc_params)
        return utils.extract_result(cr)

    def order(self, security, price, size, algo="", algo_param={}, userdata=""):
        """
        return (result, message)
        if result is None, message contains error information
        """

        r, msg = self._check_session()
        if not r: return (None, msg)

        rpc_params = {"security"    : security,
                      "price"       : price,
                      "size"        : int(size),
                      "algo"        : algo,
                      "algo_param"  : json.dumps(algo_param),
                      "user"        : self._username,
                      "userdata"    : userdata}

        cr = self._remote.call("oms.order", rpc_params)
        return utils.extract_result(cr)

    def place_order(self, security, action, price, size, algo="", algo_param={}, userdata=""):
        """
        return (result, message)
        if result is None, message contains error information
        """

        r, msg = self._check_session()
        if not r: return (None, msg)

        rpc_params = { "security"    : security,
                       "action"      : action,
                       "price"       : price,
                       "size"        : int(size),
                       "algo"        : algo,
                       "algo_param"  : json.dumps(algo_param),
                       "user"        : self._username,
                       "userdata"    : userdata}

        cr = self._remote.call("oms.place_order", rpc_params)
        return utils.extract_result(cr)

    def batch_order(self, orders, algo="", algo_param={}, userdata=""):
        """
        orders format:
            [ {"security": "000001.SZ", "action": "Buy", "price": 10.0, "size" : 100}, ... ]
        return (result, message)
        if result is None, message contains error information
        """

        if not orders or not isinstance(orders, (list, tuple)):
            return (None, "empty order")

        if isinstance(orders[0], EntrustOrder):
            tmp = []
            for o in orders:
                tmp.append({"security": o.security,
                            "price"   : o.price,
                            "size"    : int(o.size)})

            orders = tmp

        r, msg = self._check_session()
        if not r: return (None, msg)

        rpc_params = {"orders"     : orders,
                      "algo"       : algo,
                      "algo_param" : json.dumps(algo_param),
                      "user"       : self._username,
                      "userdata"   : userdata}

        cr = self._remote.call("oms.batch_order", rpc_params)
        return utils.extract_result(cr)

    def place_batch_order(self, orders, algo="", algo_param={}, userdata=""):
        """
        orders format:
            [ {"security": "000001.SZ", "action": "Buy", "price": 10.0, "size" : 100}, ... ]
        return (result, message)
        if result is None, message contains error information
        """

        if not orders or not isinstance(orders, (list, tuple)):
            return (None, "empty order")

        if isinstance(orders[0], EntrustOrder):
            tmp = []
            for o in orders:
                tmp.append({"security": o.security,
                            "action"  : o.action,
                            "price"   : o.price,
                            "size"    : int(o.size)})

            orders = tmp

        r, msg = self._check_session()
        if not r: return (None, msg)

        rpc_params = {"orders"     : orders,
                      "algo"       : algo,
                      "algo_param" : json.dumps(algo_param),
                      "user"       : self._username,
                      "userdata"   : userdata}

        cr = self._remote.call("oms.place_batch_order", rpc_params)
        return utils.extract_result(cr)

    def cancel_order(self, task_id):
        """
        return (result, message)
        if result is None, message contains error information
        """

        r, msg = self._check_session()
        if not r: return (None, msg)

        rpc_params = {"task_id": task_id}

        cr = self._remote.call("oms.cancel_order", rpc_params)
        return utils.extract_result(cr)

    def query_account(self, format=""):
        """
            return pd.dataframe
        """
        r, msg = self._check_session()
        if not r: return (None, msg)

        rpc_params = {}

        data_format = self._get_format(format, "pandas")
        if data_format == "pandas":
            rpc_params["format"] = "columnset"

        cr = self._remote.call("oms.query_account", rpc_params)

        return utils.extract_result(cr, data_format=data_format, class_name="Account")

    def query_position(self, mode="all", securities="", format=""):
        """
            securities: seperate by ","
            return pd.dataframe
        """

        r, msg = self._check_session()
        if not r: return (None, msg)

        rpc_params = {"mode"       : mode,
                      "security"   : securities}

        data_format = self._get_format(format, "pandas")
        if data_format == "pandas":
            rpc_params["format"] = "columnset"

        cr = self._remote.call("oms.query_position", rpc_params)

        return utils.extract_result(cr, data_format=data_format, class_name="Position")

    def query_net_position(self, mode="all", securities="", format=""):
        """
            securities: seperate by ","
            return pd.dataframe
        """

        r, msg = self._check_session()
        if not r: return (None, msg)

        rpc_params = {"mode"       : mode,
                      "security"   : securities}

        data_format = self._get_format(format, "pandas")
        if data_format == "pandas":
            rpc_params["format"] = "columnset"

        cr = self._remote.call("oms.query_net_position", rpc_params)

        return utils.extract_result(cr, data_format=data_format, class_name="NetPosition")

    def query_repo_contract(self, format=""):
        """
            securities: seperate by ","
            return pd.dataframe
        """

        r, msg = self._check_session()
        if not r: return (None, msg)

        rpc_params = {}

        cr = self._remote.call("oms.query_repo_contract", rpc_params)

        return utils.extract_result(cr, data_format=self._get_format(format, "pandas"), class_name="RepoContract")

    def query_task(self, task_id=-1, format=""):
        """
            task_id: -1 -- all
            return pd.dataframe
        """

        r, msg = self._check_session()
        if not r: return (None, msg)

        rpc_params = {"task_id": task_id}

        data_format = self._get_format(format, "pandas")
        if data_format == "pandas":
            rpc_params["format"] = "columnset"

        cr = self._remote.call("oms.query_task", rpc_params)

        return utils.extract_result(cr, data_format=data_format, class_name="Task")

    def query_order(self, task_id=-1, format=""):
        """
            task_id: -1 -- all
            return pd.dataframe
        """

        r, msg = self._check_session()
        if not r: return (None, msg)

        rpc_params = {"task_id": task_id}

        data_format = self._get_format(format, "pandas")
        if data_format == "pandas":
            rpc_params["format"] = "columnset"

        cr = self._remote.call("oms.query_order", rpc_params)

        return utils.extract_result(cr, data_format=data_format, class_name="Order")

    def query_trade(self, task_id=-1, format=""):
        """
            task_id: -1 -- all
            return pd.dataframe
        """

        r, msg = self._check_session()
        if not r: return (None, msg)

        rpc_params = {"task_id": task_id}

        data_format = self._get_format(format, "pandas")
        if data_format == "pandas":
            rpc_params["format"] = "columnset"

        cr = self._remote.call("oms.query_trade", rpc_params)

        return utils.extract_result(cr, data_format=data_format, class_name="Trade")

    def query_portfolio(self, format=""):
        """
            return pd.dataframe
        """

        r, msg = self._check_session()
        if not r: return (None, msg)

        rpc_params = {}

        data_format = self._get_format(format, "pandas")
        if data_format == "pandas":
            rpc_params["format"] = "columnset"

        cr = self._remote.call("pms.query_portfolio", rpc_params)

        return utils.extract_result(cr, index_column="security", data_format=data_format, class_name="NetPosition")

    def goal_portfolio(self, positions, algo="", algo_param={}, userdata=""):
        """
        positions format:
            [ {"security": "000001.SZ", "ref_price": 10.0, "size" : 100}, ...]
        return (result, message)
        if result is None, message contains error information
        """

        r, msg = self._check_session()
        if not r: return (False, msg)

        if type(positions) is pd.core.frame.DataFrame :
            tmp = []
            for i in range(0, len(positions)):
                tmp.append ({'security': positions.index[i],
                             'ref_price': float(positions['ref_price'][i]),
                              "size" : int(positions['size'][i])})
            positions = tmp

        rpc_params = {"positions"  : positions,
                      "algo"       : algo,
                      "algo_param" : json.dumps(algo_param),
                      "user"       : self._username,
                      "userdata"   : userdata}

        cr = self._remote.call("pms.goal_portfolio", rpc_params)
        return utils.extract_result(cr)

    def basket_order(self, orders, algo="", algo_param={}, userdata=""):
        """
        orders format:
            [ {"security": "000001.SZ", "ref_price": 10.0, "inc_size" : 100}, ...]
        return (result, message)
        if result is None, message contains error information
        """

        r, msg = self._check_session()
        if not r: return (False, msg)

        if type(orders) is pd.core.frame.DataFrame :
            tmp = []
            for i in range(0, len(orders)):
                tmp.append ({'security': orders.index[i],
                             'ref_price': float(orders['ref_price'][i]),
                             'inc_size' : int(orders['inc_size'][i])})
            orders = tmp

        rpc_params = {"orders": orders,
                      "algo": algo,
                      "algo_param": json.dumps(algo_param),
                      "user": self._username,
                      "userdata": userdata}

        cr = self._remote.call("pms.basket_order", rpc_params)
        return utils.extract_result(cr)

    def stop_portfolio(self):
        """
        return (result, message)
        if result is None, message contains error information
        """

        r, msg = self._check_session()
        if not r: return (False, msg)

        rpc_params = {}
    
        cr = self._remote.call("pms.stop_portfolio", rpc_params)
        return utils.extract_result(cr)

    def query_universe(self, format=""):

        r, msg  = self._check_session()
        if not r: return (None, msg)

        rpc_params = {}
        data_format = self._get_format(format, "pandas")
        if data_format == "pandas":
            rpc_params["format"] = "columnset"
        
        cr = self._remote.call("oms.query_universe", rpc_params)
        
        return utils.extract_result(cr, data_format=data_format, class_name="UniverseItem")
    
    def set_heartbeat(self, interval, timeout):
        self._remote.set_hearbeat_options(interval, timeout)
        print("heartbeat_interval =", self._remote._heartbeat_interval, ", heartbeat_timeout =",
              self._remote._heartbeat_timeout)



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import json
import random
import time

import zmq

try:
    import queue
except ImportError:
    import Queue as queue
import threading
import msgpack
import snappy
import copy

qEmpty = copy.copy(queue.Empty)


def _unpack_msgpack_snappy(str):
    if str.startswith(b'S'):
        tmp = snappy.uncompress(str[1:])
        # print "SNAPPY: ", len(str), len(tmp)
        obj = msgpack.loads(tmp, encoding='utf-8')
    elif str.startswith(b'\0'):
        obj = msgpack.loads(str[1:], encoding='utf-8')
    else:
        return None
    
    return obj


def _pack_msgpack_snappy(obj):
    # print "pack", obj
    tmp = msgpack.dumps(obj, encoding='utf-8')
    if len(tmp) > 1000:
        return b'S' + snappy.compress(tmp)
    else:
        return b'\0' + tmp


def _unpack_msgpack(str):
    return msgpack.loads(str, encoding='utf-8')


def _pack_msgpack(obj):
    return msgpack.dumps(obj, encoding='utf-8')


def _unpack_json(str):
    return json.loads(str, encoding='utf-8')


def _pack_json(obj):
    return json.dumps(obj, encoding='utf-8')


class JRpcClient(object):
    def __init__(self, data_format="msgpack_snappy"):
        self._waiter_lock = threading.Lock()
        self._waiter_map = {}

        self._should_close = False
        self._next_callid = 0
        self._send_lock = threading.Lock()
        self._callid_lock = threading.Lock()

        self._last_heartbeat_rsp_time = 0
        self._connected = False

        self.on_disconnected = None
        self.on_rpc_callback = None
        self._callback_queue = queue.Queue()
        self._call_wait_queue = queue.Queue()

        self._ctx = zmq.Context()
        self._pull_sock = self._ctx.socket(zmq.PULL)
        self._pull_sock.bind("inproc://pull_sock")
        self._push_sock = self._ctx.socket(zmq.PUSH)
        self._push_sock.connect("inproc://pull_sock")

        self._heartbeat_interval = 1
        self._heartbeat_timeout = 3

        self._addr = None
        
        if data_format == "msgpack_snappy":
            self._pack = _pack_msgpack_snappy
            self._unpack = _unpack_msgpack_snappy

        elif data_format == "msgpack":
            self._pack = _pack_msgpack
            self._unpack = _unpack_msgpack
        
        elif data_format == "json":
            self._pack = _pack_json
            self._unpack = _unpack_json
        
        else:
            assert False, "unknown data_format " + data_format
        
        t = threading.Thread(target=self._recv_run)
        t.setDaemon(True)
        t.start()
        self._recv_thread = t

        t = threading.Thread(target=self._callback_run)
        t.setDaemon(True)
        t.start()
        self._callback_thread = t

    def __del__(self):
        self.close()

    def next_callid(self):
        self._callid_lock.acquire()
        self._next_callid += 1
        callid = self._next_callid
        self._callid_lock.release()
        return callid

    def set_heartbeat_options(self, interval, timeout):
        self._heartbeat_interval = interval
        self._heartbeat_timeout = timeout

    def _recv_run(self):

        heartbeat_time = 0

        poller = zmq.Poller()
        poller.register(self._pull_sock, zmq.POLLIN)

        remote_sock = None

        while not self._should_close:

            try:
                if self._connected and time.time() - self._last_heartbeat_rsp_time > self._heartbeat_timeout:
                    self._connected = False
                    if self.on_disconnected: self._async_call(self.on_disconnected)

                if remote_sock and time.time() - heartbeat_time > self._heartbeat_interval:
                    self._send_hearbeat()
                    heartbeat_time = time.time()

                socks = dict(poller.poll(500))
                if self._pull_sock in socks and socks[self._pull_sock] == zmq.POLLIN:
                    cmd = self._pull_sock.recv()
                    if cmd == b"CONNECT":
                        # print time.ctime(), "CONNECT " + self._addr
                        if remote_sock:
                            poller.unregister(remote_sock)
                            remote_sock.close()
                            remote_sock = None

                        remote_sock = self._do_connect()

                        if remote_sock:
                            poller.register(remote_sock, zmq.POLLIN)

                    elif cmd.startswith(b"SEND:") and remote_sock:
                        # print time.ctime(), "SEND " + cmd[5:]
                        remote_sock.send(cmd[5:])

                if remote_sock and remote_sock in socks and socks[remote_sock] == zmq.POLLIN:
                    data = remote_sock.recv()
                    if data:
                        # if not data.find("heartbeat"):
                        #    print time.ctime(), "RECV", data
                        self._on_data_arrived(data)

            except zmq.error.Again as e:
                # print "RECV timeout: ", e
                pass
            except Exception as e:
                print("_recv_run:", e)

    def _callback_run(self):
        while not self._should_close:
            try:
                r = self._callback_queue.get(timeout=1)
                if r:
                    r()
            except qEmpty as e:
                pass
            except TypeError as e:
                if str(e) == "'NoneType' object is not callable":
                    pass
                else:
                    print("_callback_run {}".format(r), type(e), e)
            except Exception as e:
                print("_callback_run {}".format(r), type(e), e)

    def _async_call(self, func):
        self._callback_queue.put(func)

    def _send_request(self, json):

        try:
            self._send_lock.acquire()
            self._push_sock.send(b"SEND:" + json)

        finally:
            self._send_lock.release()

    def connect(self, addr):
        self._addr = addr
        self._push_sock.send_string('CONNECT', encoding='utf-8')

    def _do_connect(self):

        client_id = str(random.randint(1000000, 100000000))

        socket = self._ctx.socket(zmq.DEALER)
        identity = (client_id) + '$' + str(random.randint(1000000, 1000000000))
        identity = identity.encode('utf-8')
        socket.setsockopt(zmq.IDENTITY, identity)
        socket.setsockopt(zmq.RCVTIMEO, 500)
        socket.setsockopt(zmq.SNDTIMEO, 500)
        socket.setsockopt(zmq.LINGER, 0)
        socket.connect(self._addr)

        return socket

    def close(self):
        self._should_close = True
        self._callback_thread.join()
        self._recv_thread.join()

    def _on_data_arrived(self, str):
        try:
            msg = self._unpack(str)
            # print "RECV", msg

            if not msg:
                print("wrong message format")
                return

            if 'method' in msg and msg['method'] == '.sys.heartbeat':
                self._last_heartbeat_rsp_time = time.time()
                if not self._connected:
                    self._connected = True
                    if self.on_connected:
                        self._async_call(self.on_connected)
                
                # Let user has a chance to check message in .sys.heartbeat
                if 'result' in msg and self.on_rpc_callback:
                    self._async_call(lambda: self.on_rpc_callback(msg['method'], msg['result']))

            elif 'id' in msg and msg['id']:
                
                # Call result
                id = int(msg['id'])
                
                if self._waiter_lock.acquire():
                    if id in self._waiter_map:
                        q = self._waiter_map[id]
                        if q: q.put(msg)
                    self._waiter_lock.release()
            else:
                # Notification message
                if 'method' in msg and 'result' in msg and self.on_rpc_callback:
                    self._async_call(lambda: self.on_rpc_callback(msg['method'], msg['result']))
        
        except Exception as e:
            print("_on_data_arrived:", e)
            pass
    
    def _send_hearbeat(self):
        msg = {'jsonrpc': '2.0',
               'method': '.sys.heartbeat',
               'params': {'time': time.time()},
               'id': str(self.next_callid())}
        json_str = self._pack(msg)
        self._send_request(json_str)

    def _alloc_wait_queue(self):
        self._waiter_lock.acquire()
        if self._call_wait_queue:
            q = self._call_wait_queue
            self._call_wait_queue = None
        else:
            q = queue.Queue()
        self._waiter_lock.release()
        return q

    def _free_wait_queue(self, q):
        self._waiter_lock.acquire()
        if not self._call_wait_queue:
            self._call_wait_queue = q
        else:
            del q
        self._waiter_lock.release()

    def call(self, method, params, timeout=6):
        # print "call", method, params, timeout
        callid = self.next_callid()
        if timeout:
            q = self._alloc_wait_queue()

            self._waiter_lock.acquire()
            self._waiter_map[callid] = q
            self._waiter_lock.release()
        
        msg = {'jsonrpc': '2.0',
               'method': method,
               'params': params,
               'id': str(callid)}

        # print "SEND", msg
        json_str = self._pack(msg)
        self._send_request(json_str)
        
        if timeout:
            ret = {}
            try:
                r = q.get(timeout=timeout)
                q.task_done()
            except qEmpty:
                r = None

            self._waiter_lock.acquire()
            self._waiter_map[callid] = None
            self._waiter_lock.release()
            self._free_wait_queue(q)

            if r:
                if 'result' in r:
                    ret['result'] = r['result']

                if 'error' in r:
                    ret['error'] = r['error']

            return ret if ret else {'error': {'error': -1, 'message': "timeout"}}
        else:
            return {'result': True}



# encoding: utf-8

from __future__ import print_function
from __future__ import absolute_import
import time

import numpy as np

from jaqs.trade import common
from jaqs.trade import EventDrivenStrategy
from jaqs.data import RemoteDataService
from jaqs.data.basic import Bar, Quote
from jaqs.trade import model
from jaqs.trade import EventLiveTradeInstance
from jaqs.trade import EventBacktestInstance
from jaqs.trade import RealTimeTradeApi, BacktestTradeApi
from jaqs.trade import PortfolioManager
import jaqs.trade.analyze as ana
import jaqs.util as jutil

from config_path import DATA_CONFIG_PATH, TRADE_CONFIG_PATH
data_config = jutil.read_json(DATA_CONFIG_PATH)
trade_config = jutil.read_json(TRADE_CONFIG_PATH)

result_dir_path = '../output/double_ma'


class DoubleMaStrategy(EventDrivenStrategy):
    """"""
    def __init__(self):
        super(DoubleMaStrategy, self).__init__()
        self.symbol = ''
        
        self.fast_ma_len = 13
        self.slow_ma_len = 23
        
        self.window_count = 0
        self.window = self.slow_ma_len + 1
        
        self.price_arr = np.zeros(self.window)
        self.fast_ma = 0
        self.slow_ma = 0
        self.pos = 0
        
        self.buy_size_unit = 1
        self.output = True
    
    def init_from_config(self, props):
        super(DoubleMaStrategy, self).init_from_config(props)
        self.symbol = props.get('symbol')
        self.init_balance = props.get('init_balance')
    
    def buy(self, quote, size=1):
        if isinstance(quote, Quote):
            ref_price = (quote.bidprice1 + quote.askprice1) / 2.0
        else:
            ref_price = quote.close
            
        task_id, msg = self.ctx.trade_api.place_order(quote.symbol, common.ORDER_ACTION.BUY, ref_price + 3, self.buy_size_unit * size)
        if (task_id is None) or (task_id == 0):
            print("place_order FAILED! msg = {}".format(msg))
    
    def sell(self, quote, size=1):
        if isinstance(quote, Quote):
            ref_price = (quote.bidprice1 + quote.askprice1) / 2.0
        else:
            ref_price = quote.close
    
        task_id, msg = self.ctx.trade_api.place_order(quote.symbol, common.ORDER_ACTION.SHORT, ref_price - 3, self.buy_size_unit * size)
        if (task_id is None) or (task_id == 0):
            print("place_order FAILED! msg = {}".format(msg))
    
    """
    `on_tick` accepts single Quote, while `on_bar` accepts
    'on_tick' is used for real-time trading, while 'on_bar' is used for backtest
    """
    def on_tick(self, quote):
        # 'quote' can be:
        #     - Quote (at real-time trading, data comes as ticks)
        #     - Bar (backtest on minutely or daily frequency)
        # we use 'isinstance' to check whether 'quote' is Quote or Bar
        if isinstance(quote, Quote):
            bid, ask = quote.bidprice1, quote.askprice1
            if bid > 0 and ask > 0:
                mid = (quote.bidprice1 + quote.askprice1) / 2.0
            else:
                # price reached high_limit or low_limit, we do not trade
                return
        else:
            mid = quote.close
        
        self.price_arr[0: self.window - 1] = self.price_arr[1: self.window]
        self.price_arr[-1] = mid
        self.window_count += 1
    
        if self.window_count <= self.window:
            return
    
        self.fast_ma = np.mean(self.price_arr[-self.fast_ma_len - 1:])
        self.slow_ma = np.mean(self.price_arr[-self.slow_ma_len - 1:])
    
        print(quote)
        print("Fast MA = {:.2f}     Slow MA = {:.2f}".format(self.fast_ma, self.slow_ma))
        if self.fast_ma > self.slow_ma:
            if self.pos == 0:
                self.buy(quote, 1)
        
            elif self.pos < 0:
                self.buy(quote, 2)
    
        elif self.fast_ma < self.slow_ma:
            if self.pos == 0:
                self.sell(quote, 1)
            elif self.pos > 0:
                self.sell(quote, 2)

    def on_bar(self, quote_dic):
        quote = quote_dic.get(self.symbol)
        self.on_tick(quote)

    def on_trade(self, ind):
        print("\nStrategy on trade: ")
        print(ind)
        self.pos = self.ctx.pm.get_pos(self.symbol)

    def on_order_status(self, ind):
        if self.output:
            print("\nStrategy on order status: ")
            print(ind)

    def on_task_status(self, ind):
        if self.output:
            print("\nStrategy on task ind: ")
            print(ind)


def test_backtest():
    props = {"symbol": "rb1710.SHF",
             "start_date": 20170710,
             "end_date": 20170730,
             "bar_type": "1M",  # '1d'
             "init_balance": 2e4}

    tapi = BacktestTradeApi()
    ins = EventBacktestInstance()
        
    props.update(data_config)
    props.update(trade_config)
    
    ds = RemoteDataService()
    strat = DoubleMaStrategy()
    pm = PortfolioManager()
    
    context = model.Context(data_api=ds, trade_api=tapi, instance=ins,
                            strategy=strat, pm=pm)
    
    ins.init_from_config(props)

    ins.run()
    ins.save_results(result_dir_path)
    do_analyze()


def test_livetrade():
    props = {'symbol': 'rb1801.SHF',
             'strategy_no': 1044}
    tapi = RealTimeTradeApi(trade_config)
    ins = EventLiveTradeInstance()
    
    props.update(data_config)
    props.update(trade_config)
    
    ds = RemoteDataService()
    strat = DoubleMaStrategy()
    pm = PortfolioManager()
    
    context = model.Context(data_api=ds, trade_api=tapi, instance=ins,
                            strategy=strat, pm=pm)
    
    ins.init_from_config(props)
    ds.subscribe(props['symbol'])
    
    ins.run()
    time.sleep(3)
    ins.stop()
    ins.save_results(result_dir_path)
    do_analyze()


def do_analyze():
    from jaqs.trade.analyze.analyze import TradeRecordEmptyError
    ta = ana.EventAnalyzer()
    
    ds = RemoteDataService()
    ds.init_from_config(data_config)
    
    try:
        ta.initialize(data_server_=ds, file_folder=result_dir_path)
        
        ta.do_analyze(result_dir=result_dir_path, selected_sec=[])
        
    except TradeRecordEmptyError:
        pass


if __name__ == "__main__":
    test_backtest()
    test_livetrade()



# encoding: utf-8

from __future__ import print_function
import numpy as np
import pandas as pd
import jaqs.util as jutil


def test_read_save_pickle():
    fp = '../../output/tests/test_read_save_pickle.pic'
    d = {'a': 1.0, 'b': 2, 'c': True, 'd': list()}
    jutil.save_pickle(d, fp)
    
    d2 = jutil.load_pickle(fp)
    assert d2['b'] == 2
    
    d3 = jutil.load_pickle('a_non_exits_file_blabla.pic')
    assert d3 is None


def test_read_save_json():
    fp = '../../output/tests/test_read_save_pickle.pic'
    d = {'a': 1.0, 'b': 2, 'c': True, 'd': list()}
    jutil.save_json(d, fp)
    
    d2 = jutil.read_json(fp)
    assert d2['b'] == 2
    
    d3 = jutil.read_json('a_non_exits_file_blabla.pic')
    assert d3 == dict()


def test_combine_date_time():
    date = 20170101
    time = 145930
    assert jutil.combine_date_time(date, time) == 20170101145930

    a = np.arange(20170101, 20170101+500, dtype=np.int64)
    b = np.arange(93000, 93000+500, dtype=np.int64)
    assert np.all(jutil.combine_date_time(a, b) == a * 1000000 + b)


def test_seq_gen():
    seq_gen = jutil.SequenceGenerator()
    for i in range(1, 100):
        assert seq_gen.get_next('a') == i


def test_timer():
    import time
    timer = jutil.SimpleTimer()
    timer.tick('start')
    time.sleep(1.5)
    timer.tick('e1')
    time.sleep(0.5)
    timer.tick('e2')


def test_pdutil():
    df = pd.DataFrame(np.random.rand(4, 20))
    df.iloc[1, 2] = np.nan
    df.iloc[3, 4] = np.nan
    df.iloc[1, 4] = np.nan
    assert df.isnull().sum().sum() == 3
    df.iloc[2, 11] = np.inf
    df.iloc[2, 12] = -np.inf
    assert df.isnull().sum().sum() == 3
    df2 = jutil.fillinf(df)
    assert df2.isnull().sum().sum() == 5

    res_q = jutil.to_quantile(df, 5, axis=1)
    
    df3 = df.copy()
    df3['group'] = ['a', 'a', 'b', 'a']
    
    dic = jutil.group_df_to_dict(df3, by='group')
    assert set(list(dic.keys())) == {'a', 'b'}
    

def test_dtutil():
    import datetime
    date = 20170105
    dt = jutil.convert_int_to_datetime(date)
    
    assert jutil.shift(date, 2) == 20170119
    assert jutil.convert_datetime_to_int(jutil.shift(dt, 2)) == 20170119
    
    t = 145539
    dt = jutil.combine_date_time(date, t)
    assert jutil.split_date_time(dt) == (date, t)
    
    assert jutil.get_next_period_day(date, 'day', 1, 1) == 20170109
    assert jutil.get_next_period_day(date, 'week', 2, 0) == 20170116
    assert jutil.get_next_period_day(date, 'month', 1, 0) == 20170201
    
    date = 20170808
    assert jutil.get_next_period_day(20170831, 'day', extra_offset=1) == 20170904
    assert jutil.get_next_period_day(20170831, 'day', n=2, extra_offset=0) == 20170904
    assert jutil.get_next_period_day(20170831, 'day', n=7, extra_offset=0) == 20170911
    assert jutil.get_next_period_day(20170831, 'week', extra_offset=1) == 20170905
    assert jutil.get_next_period_day(20170831, 'month', extra_offset=0) == 20170901
    
    monthly = 20170101
    while monthly < 20180301:
        monthly = jutil.get_next_period_day(monthly, 'month', extra_offset=0)
        assert datetime.datetime.strptime(str(monthly), "%Y%m%d").weekday() < 5


def test_rank_percentile():
    df = pd.DataFrame(np.random.rand(500, 3000))
    res1 = jutil.rank_with_mask(df, axis=1, mask=None, normalize=False)
    res2 = jutil.rank_with_mask(df, axis=1, mask=None, normalize=True)
    print
    
    #assert np.nanmean(val[res == 1].values.flatten()) < 0.11
    #
    #val = pd.DataFrame(np.random.rand(1000, 100))
    #expr = parser.parse('Ts_Quantile(val, 500, 12)')
    #res = parser.evaluate({'val': val})
    #assert np.nanmean(val[res == 1].values.flatten()) < 0.11


def test_io():
    folder_relative = '../output/test/test_file_io'
    folder = jutil.join_relative_path(folder_relative)
    fp = jutil.join_relative_path(folder_relative+'/file.postfix')
    
    jutil.create_dir(fp)
    jutil.create_dir(folder)
    

def test_base64():
    import matplotlib
    matplotlib.use('Agg')
    
    import matplotlib.pyplot as plt
    fig = plt.figure()
    plt.plot(range(10))
    res = jutil.fig2base64(fig, 'png')
    res = jutil.fig2base64(fig, 'pdf')


def test_is_numeric():
    from jaqs.util import is_numeric
    
    NUMERIC = [True, 1, -1, 1.0, 1+1j]
    NOT_NUMERIC = [object(), 'string', u'unicode', None]
    
    def test_is_numeric(self):
        for x in self.NUMERIC:
            for y in (x, [x], [x] * 2):
                for z in (y, np.array(y)):
                    self.assertTrue(is_numeric(z))
        for x in self.NOT_NUMERIC:
            for y in (x, [x], [x] * 2):
                for z in (y, np.array(y)):
                    self.assertFalse(is_numeric(z))
        for kind, dtypes in np.sctypes.items():
            if kind != 'others':
                for dtype in dtypes:
                    self.assertTrue(is_numeric(np.array([0], dtype=dtype)))


if __name__ == "__main__":
    import time
    t_start = time.time()
    
    g = globals()
    g = {k: v for k, v in g.items() if k.startswith('test_') and callable(v)}
    
    for test_name, test_func in g.items():
        print("\n==========\nTesting {:s}...".format(test_name))
        test_func()
    print("Test Complete.")
    
    t3 = time.time() - t_start
    print("\n\n\nTime lapsed in total: {:.1f}".format(t3))



# encoding: utf-8

from __future__ import absolute_import, division, print_function

from jaqs.trade import model


def test_model():
    model.StockSelector()
    model.SimpleCostModel()
    model.AlphaContext()
    model.FactorRiskModel()
    model.FactorSignalModel()
    


# encoding: UTF-8

from __future__ import print_function
from jaqs.data import DataApi
from jaqs.trade import common
import jaqs.util as jutil

from config_path import DATA_CONFIG_PATH
data_config = jutil.read_json(DATA_CONFIG_PATH)


def test_data_api():
    address = data_config.get("remote.data.address", None)
    username = data_config.get("remote.data.username", None)
    password = data_config.get("remote.data.password", None)
    if address is None or username is None or password is None:
        raise ValueError("no data service config available!")
    
    api = DataApi(address, use_jrpc=False)
    login_msg = api.login(username=username, password=password)
    print(login_msg)
    
    daily, msg = api.daily(symbol="600030.SH,000002.SZ", start_date=20170103, end_date=20170708,
                           fields="open,high,low,close,volume,last,trade_date,settle")
    daily2, msg2 = api.daily(symbol="600030.SH", start_date=20170103, end_date=20170708,
                             fields="open,high,low,close,volume,last,trade_date,settle")
    # err_code, err_msg = msg.split(',')
    assert msg == '0,'
    assert msg2 == '0,'
    assert daily.shape == (248, 9)
    assert daily2.shape == (124, 9)
    
    df, msg = api.bar(symbol="600030.SH", trade_date=20170904, freq=common.QUOTE_TYPE.MIN, start_time=90000, end_time=150000)
    print(df.columns)
    assert df.shape == (240, 15)
    
    print("test passed")
    

if __name__ == "__main__":
    test_data_api()



# encoding: utf-8
"""
Classes defined in marketdata module represent different market data, including:
- Bar
- Quote

"""


class Bar(object):
    """
    A Bar is a summary of information of price and volume
    during a certain length of time span.
    
    Attributes
    ----------
    symbol : str
    open : float
        The first price in the time span.
    high : float
        The highest price in the time span.
    low  : float
        The lowest price in the time span.
    close : float
        The last price in the time span.
    vwap : float
        The volume-weighted average price of the time span.
    volume : int
        Trading volume within or at then end of the time span.
    oi : int
        Open interest at the end of the time span.
    trade_date : int
        %YY%mm%dd
    time : int
        %HH%mm%ss
    
    """
    def __init__(self):
        self.symbol = ""
        self.open = 0.
        self.close = 0.
        self.high = 0.
        self.low = 0.
        self.volume = 0
        self.oi = 0
        self.vwap = 0.
        
        self.trade_date = 0
        self.time = 0
        self._bar_keys = ['open', 'close', 'high', 'low', 'vwap',
                          'volume', 'oi',
                          'trade_date', 'time']
    
    @classmethod
    def create_from_df(cls, df):
        """
        Create a list of Bars from a DataFrame.
        
        Parameters
        ----------
        df : pd.DataFrame
            Index does not matter. Each row contains information of a Bar.

        Returns
        -------
        bar_list : list of Bar

        """
        bar_list = []
        for _, row in df.iterrows():
            dic = row.to_dict()
            bar = Bar.create_from_dict(dic)
            bar_list.append(bar)
        return bar_list

    @classmethod
    def create_from_dict(cls, dic):
        """
        Create a list of Bars from a dic.
        
        Parameters
        ----------
        dic : dict
            Keys are attribute names of Bar, values are corresponding values.

        Returns
        -------
        bar : Bar

        """
        bar = cls()
        bar.__dict__.update(dic)
        return bar

    def __repr__(self):
        return "{0.trade_date:8d}-{0.time:6d} " \
               "{0.volume:13.2f} of " \
               "{0.symbol:10s}@{0.close:.3f}".format(self)

    def __str__(self):
        return self.__repr__()


class Quote(object):
    """
    Quote represents a snapshot of price and volume information.
    
    Attributes
    ----------
    open : float
        The daily open price.
    high : float
        The daily highest price.
    low  : float
        The daily lowest price.
    close : float
    volume : int
        Trading volume till now.
    turnover : int
        Trading turnover till now.
    oi : int
        Open interest till now.
    trade_date : int
        %YY%mm%dd
    date : int
        %YY%mm%dd, natural date.
    time : int
        %HH%mm%ss
    last : float
        Latest trading price.
    askprice1 : float
        Best ask price.
    bidprice1 : float
        Best bid price.
    askvolume1 : float
        Sum of tradable volume of all orders at the best ask price.
    bidvolume1 : float
        Sum of tradable volume of all orders at the best bid price.
    
    
    """
    def __init__(self):
        self.trade_date = 0
        self.date = 0
        self.time = 0
        
        self.open = 0.
        self.close = 0.
        self.high = 0.
        self.low = 0.
        self.vwap = 0.
        
        self.settle = 0.0
        
        self.volume = 0
        self.turnover = 0
        self.oi = 0
        
        self.preclose = 0.0
        self.presettle = 0.0
        self.preoi = 0
        
        self.last = 0.0
        self.bidprice1 = 0.0
        self.bidprice2 = 0.0
        self.bidprice3 = 0.0
        self.bidprice4 = 0.0
        self.bidprice5 = 0.0
        self.askprice1 = 0.0
        self.askprice2 = 0.0
        self.askprice3 = 0.0
        self.askprice4 = 0.0
        self.askprice5 = 0.0

        self.bidvolume1 = 0.0
        self.bidvolume2 = 0.0
        self.bidvolume3 = 0.0
        self.bidvolume4 = 0.0
        self.bidvolume5 = 0.0
        self.askvolume1 = 0.0
        self.askvolume2 = 0.0
        self.askvolume3 = 0.0
        self.askvolume4 = 0.0
        self.askvolume5 = 0.0
        
        self.limit_up = 0.0
        self.limit_down = 0.0

    @classmethod
    def create_from_dict(cls, dic):
        quote = cls()
        quote.__dict__.update(dic)
        return quote

    def __repr__(self):
        return ("{0.symbol:s}  {0.trade_date:8d}-{0.time:6d}      "
                " (BID) {0.bidvolume1:6.0f}@{0.bidprice1:6.2f}"
                " | "
                "{0.askprice1:6.2f}@{0.askvolume1:<6.0f} (ASK)".format(self))

    def __str__(self):
        return self.__repr__()




from setuptools import setup, find_packages
from jaqs import __version__ as ver
import codecs
import os


def read(fname):
    return codecs.open(os.path.join(os.path.dirname(__file__), fname), 'r', encoding='utf-8').read()


def readme():
    with codecs.open('README.rst', 'r', encoding='utf-8') as f:
        return f.read()


def read_install_requires():
    with codecs.open('requirements.txt', 'r', encoding='utf-8') as f:
        res = f.readlines()
    res = list(map(lambda s: s.replace('\n', ''), res))
    return res

setup(
    # Install data files specified in MANIFEST.in file.
    include_package_data=True,
    #package_data={'': ['*.json', '*.css', '*.html']},
    # Package Information
    name='jaqs',
    url='https://github.com/quantOS-org/JAQS',
    version=ver,
        license='Apache 2.0',
    # information
    description='Open source quantitative research&trading framework.',
    long_description=readme(),
    keywords="quantiatitive trading research finance",
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "Intended Audience :: Education",
        "Intended Audience :: End Users/Desktop",
        "Intended Audience :: Financial and Insurance Industry",
        "Intended Audience :: Information Technology",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: Apache Software License",
        "Natural Language :: Chinese (Simplified)",
        "Natural Language :: English",
        "Operating System :: Microsoft :: Windows",
        "Operating System :: Unix",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3.6",
    ],
    # install
    install_requires=read_install_requires(),
    packages=find_packages(),
    # author
    author='quantOS'
)



# encoding: utf-8
"""
jaqs.data.basic module includes basic definitions of basic data types
that are used across jaqs package.

Basic data types:
- `Instrument`: An Instrument represents a specific financial contract.
- `InstManager`: InstManager query information of instruments from data server and store them.
- `Bar`: A Bar is a summary of information of price and volume during a certain length of time span.
- `Quote`: Quote represents a snapshot of price and volume information.
- `Order`: Basic order class.
- `OrderStatusInd`: OrderStatusInd is a indication of status change of an order.
- `Task`: Task is a high-level trading target, which may contain many orders.
- `Position`: Basic position class.
- `Trade`: Trade represents fill/partial fill of an order.
- `TradeInd`: TaskInd is a indication of status change of a task.
- `TradeStat`: TradeStat stores statistics of trading of a certain symbol.

"""
from .marketdata import *
from .order import *
from .trade import *
from .position import *
from .instrument import *
